// Copyright 2001-2002 Omni Development, Inc.  All rights reserved.
//
// This software may only be used and reproduced according to the
// terms in the file OmniSourceLicense.html, which should be
// distributed with this project and can also be found at
// http://www.omnigroup.com/DeveloperResources/OmniSourceLicense.html.

#import "OARegExFindPattern.h"

#import <Foundation/Foundation.h>
#import <OmniBase/OmniBase.h>
#import <OmniFoundation/OmniFoundation.h>

RCS_ID("$Header: /Network/Source/CVS/OmniGroup/Frameworks/OmniAppKit/OARegExFindPattern.m,v 1.7 2002/03/09 01:53:53 kc Exp $")

@implementation OARegExFindPattern

- initWithString:(NSString *)aString selectedSubexpression:(int)subexpression backwards:(BOOL)backwards;
{
    [super init];
    regularExpression = [[OFRegularExpression alloc] initWithString:aString];
    selectedSubexpression = subexpression;
    isBackwards = backwards;
    return self;
}

- (void)setReplacementString:(NSString *)aString;
{
    if (aString != replacementString) {
        [replacementString release];
        replacementString = [aString retain];
    }
}

- (void)dealloc;
{
    [regularExpression release];
    [lastMatch release];
    [replacementString release];
    [super dealloc];
}

//
// OAFindPattern protocol
//

- (BOOL)findInString:(NSString *)aString foundRange:(NSRangePointer)rangePtr;
{
    OFRegularExpressionMatch *match;
    
    [lastMatch release];
    lastMatch = nil;
    
    if (aString == nil)
        return NO;

    if (!(match = [regularExpression matchInString:aString]))
        return NO;
        
    if (isBackwards) {
        OFRegularExpressionMatch *next;

        while ((next = [match nextMatch]))
            match = next;
    }
        
    if (rangePtr != NULL) {
        if (selectedSubexpression == SELECT_FULL_EXPRESSION)
            *rangePtr = [match matchRange];
        else
            *rangePtr = [match rangeOfSubexpressionAtIndex:selectedSubexpression];
    }
    
    lastMatch = [match retain];
    return YES;
}

- (BOOL)findInRange:(NSRange)range ofString:(NSString *)aString foundRange:(NSRangePointer)rangePtr;
{
    BOOL result;
    
    if (aString == nil)
        return NO;

    result = [self findInString:[aString substringWithRange:range] foundRange:rangePtr];
    if (rangePtr != NULL)
        rangePtr->location += range.location;
    return result;
}

- (NSString *)replacementStringForLastFind;
{
    OFStringScanner *scanner;
    NSMutableString *interpolatedString = [NSMutableString string];
    
    scanner = [[OFStringScanner alloc] initWithString:replacementString];
    while (scannerHasData(scanner)) {
        int subexpressionIndex = 0;
        BOOL readNumber = NO;
        unichar c;

        [interpolatedString appendString:[scanner readFullTokenWithDelimiterCharacter:'\\']];
        if (scannerReadCharacter(scanner) != '\\')
            break;
        
        while ((c = scannerPeekCharacter(scanner)) && (c >= '0') && (c <= '9')) {
            scannerSkipPeekedCharacter(scanner);
            subexpressionIndex *= 10;
            subexpressionIndex += (c - '0');
            readNumber = YES;
        }
        if (readNumber && subexpressionIndex <= [regularExpression subexpressionCount]) {
            NSString *subString;
            
            if (subexpressionIndex)
                subString = [lastMatch subexpressionAtIndex:(subexpressionIndex - 1)];
            else	
                subString = [lastMatch matchString];
            [interpolatedString appendString:subString];
        } else if (c == '\\') {
            [interpolatedString appendString:@"\\"];
        }
    }
    [scanner release];
    return interpolatedString;
}

@end
